"""Argument unparser."""

import logging
import typing as t

_LOG = logging.getLogger(__name__)


def treat_as_option_with_no_value(value: t.Any) -> bool:
    """Determine if an option should be treated as a flag with no value."""
    return value is True


def option_should_be_skipped(value: t.Any) -> bool:
    """Determine if an option should be skipped."""
    return value is False or value is None


class ArgumentUnparser:
    """For performing reverse operation to what argparse.ArgumentParser does."""

    # pylint: disable = too-many-arguments, too-many-positional-arguments
    def __init__(
            self, short_opt: str = '-', long_opt: str = '--', opt_value: str = '=',
            begin_delim: str = '"', end_delim: str = '"') -> None:
        assert isinstance(short_opt, str)
        assert isinstance(long_opt, str)
        assert isinstance(opt_value, str)
        assert isinstance(begin_delim, str)
        assert isinstance(end_delim, str)

        self._short_opt = short_opt
        self._long_opt = long_opt
        self._opt_value = opt_value
        self._begin_delim = begin_delim
        self._end_delim = end_delim

    def unparse_arg(self, arg: t.Any) -> str:  # pylint: disable = no-self-use
        """Convert an object into a string that can be used as a command-line argument."""
        if not isinstance(arg, str):
            arg = str(arg)
        if ' ' in arg:
            arg = repr(arg)
        if not arg:
            arg = '""'
        return arg

    def unparse_args_to_list(self, arguments: t.Sequence[t.Any]) -> list[str]:
        """Convert list of objects to a list of command-line args."""
        unparsed_list = []
        for arg in arguments:
            unparsed_list.append(self.unparse_arg(arg))
        _LOG.debug('%s: unparsed args to %s', self, unparsed_list)
        return unparsed_list

    def unparse_args(self, arguments: t.Sequence[t.Any]) -> str:
        """Convert list to string of command-line args."""
        unparsed = ' '.join(self.unparse_args_to_list(arguments))
        _LOG.debug('%s: converted unparsed args to string "%s"', self, unparsed)
        return unparsed

    def unparse_option_to_list(self, key: str, value: t.Any) -> list[str]:
        """Convert a key-value pair into a short list to be used as a command-line option."""
        if option_should_be_skipped(value):
            return []
        unparsed_key = f'{self._long_opt if len(key) > 1 else self._short_opt}{key}'
        if treat_as_option_with_no_value(value):
            return [unparsed_key]
        unparsed_value = self.unparse_arg(value)
        if self._opt_value == ' ':
            return [unparsed_key, unparsed_value]
        return [f'{unparsed_key}{self._opt_value}{unparsed_value}']

    def unparse_option(self, key: str, value: t.Any) -> str:
        """Convert a key-value pair into a string that can be used as a command-line option."""
        if option_should_be_skipped(value):
            return ''
        unparsed = ' '.join(self.unparse_option_to_list(key, value))
        return unparsed

    def unparse_options_to_list(self, options: t.Mapping[str, t.Any]) -> list[str]:
        """Convert dictionary to a list of command-line args."""
        unparsed_list: t.List[str] = []
        for key, value in options.items():
            if option_should_be_skipped(value):
                continue
            unparsed_list += self.unparse_option_to_list(key, value)
        _LOG.debug('%s: unparsed options to %s', self, unparsed_list)
        return unparsed_list

    def unparse_options(self, options: t.Mapping[str, t.Any]) -> str:
        """Convert dictionary to string of command-line args."""
        unparsed = ' '.join(self.unparse_options_to_list(options))
        _LOG.debug('%s: converted unparsed options to string "%s"', self, unparsed)
        return unparsed

    def unparse_options_and_args_to_list(
            self, options: t.Mapping[str, t.Any], arguments: t.Sequence[t.Any]) -> list[str]:
        """Convert dictionary and list to a list of command-line args."""
        unparsed_options = [] if options is None else self.unparse_options_to_list(options)
        unparsed_args = [] if arguments is None else self.unparse_args_to_list(arguments)
        return unparsed_options + unparsed_args

    def unparse_options_and_args(
            self, options: t.Mapping[str, t.Any], arguments: t.Sequence[t.Any]) -> str:
        """Convert dictionary and list to string of command-line args."""
        return ' '.join(self.unparse_options_and_args_to_list(options, arguments))

    def unparse_to_list(self, *args, **kwargs) -> list:
        """Unparse given args as command-line arguments and kwargs as command-line options.

        Output is a list of strings.

        This process is a reverse of what built-in argparse module does with parse_args() method.
        """
        return self.unparse_options_and_args_to_list(kwargs, args)

    def unparse(self, *args, **kwargs) -> str:
        """Unparse given args as command-line arguments and kwargs as command-line options.

        This process is a reverse of what built-in argparse module does with parse_args() method.
        """
        return self.unparse_options_and_args(kwargs, args)

    def __str__(self):
        return f'{type(self).__name__}(short="{self._short_opt}",long="{self._long_opt}",' \
            f'value="{self._opt_value}")'



"""Initialization of tests of argunparse package."""

import logging

import boilerplates.logging


class TestsLogging(boilerplates.logging.Logging):
    """Logging configuration for tests."""

    packages = ['argunparse']


TestsLogging.configure()
logging.getLogger('argunparse.argument_unparser').setLevel(logging.INFO)




"""Tests for ArgumentUnparser class."""

import itertools
import logging
import unittest

from argunparse.argument_unparser import ArgumentUnparser
from .examples import \
    OPTIONS, OPTIONS_VARIANTS, OPTIONS_SKIPPED, OPTIONS_SKIPPED_VARIANTS, \
    ARGUMENTS, ARGUMENTS_VARIANTS, OPTIONS_AND_ARGUMENTS_VARIANTS

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_construct(self):
        unparser = ArgumentUnparser()
        self.assertIsInstance(unparser, ArgumentUnparser)

    def test_option(self):
        unparser = ArgumentUnparser()
        for reference, option in OPTIONS.items():
            with self.subTest(option=option):
                key, value = list(*itertools.chain(option.items()))
                list_result = unparser.unparse_option_to_list(key, value)
                self.assertListEqual([reference], list_result)
                result = unparser.unparse_option(key, value)
                self.assertEqual(reference, result)

    def test_option_skipped(self):
        unparser = ArgumentUnparser()
        for option in OPTIONS_SKIPPED.items():
            with self.subTest(option=option):
                key, value = option
                list_result = unparser.unparse_option_to_list(key, value)
                self.assertListEqual(list_result, [])
                result = unparser.unparse_option(key, value)
                self.assertEqual(result, '')

    def test_option_space(self):
        unparser = ArgumentUnparser(opt_value=' ')
        for reference, option in OPTIONS.items():
            with self.subTest(option=option):
                key, value = list(*itertools.chain(option.items()))
                list_result = unparser.unparse_option_to_list(key, value)
                self.assertListEqual(reference.split('='), list_result)
                result = unparser.unparse_option(key, value)
                self.assertEqual(reference.replace('=', ' '), result)

    def test_options(self):
        _LOG.debug('testing %i option variants...',
                   len(OPTIONS_VARIANTS) + len(OPTIONS_SKIPPED_VARIANTS))
        unparser = ArgumentUnparser()
        for reference, options in itertools.chain(OPTIONS_VARIANTS, OPTIONS_SKIPPED_VARIANTS):
            with self.subTest(options=options):
                list_result = unparser.unparse_options_to_list(options)
                self.assertListEqual(reference, list_result)
                result = unparser.unparse_options(options)
                self.assertEqual(' '.join(reference), result)

    def test_options_space(self):
        _LOG.debug('testing %i option variants...',
                   len(OPTIONS_VARIANTS) + len(OPTIONS_SKIPPED_VARIANTS))
        unparser = ArgumentUnparser(opt_value=' ')
        for reference, options in itertools.chain(OPTIONS_VARIANTS, OPTIONS_SKIPPED_VARIANTS):
            with self.subTest(options=options):
                result = unparser.unparse_options(options)
                self.assertEqual(' '.join(reference).replace('=', ' '), result)
                list_result = unparser.unparse_options_to_list(options)
                fixed_reference = list(itertools.chain.from_iterable(
                    [_.split('=') if '=' in _ else [_] for _ in reference]))
                self.assertListEqual(fixed_reference, list_result)

    def test_arg(self):
        unparser = ArgumentUnparser()
        for reference, arg in ARGUMENTS:
            with self.subTest(arg=arg):
                result = unparser.unparse_arg(arg)
                self.assertEqual(reference, result)

    def test_args(self):
        unparser = ArgumentUnparser()
        for reference, args in ARGUMENTS_VARIANTS:
            with self.subTest(args=args):
                list_result = unparser.unparse_args_to_list(args)
                self.assertListEqual(reference, list_result)
                result = unparser.unparse_args(args)
                self.assertEqual(' '.join(reference), result)

    def test_options_and_args(self):
        unparser = ArgumentUnparser()
        for (reference_options, options), (reference_args, args) in OPTIONS_AND_ARGUMENTS_VARIANTS:
            with self.subTest(options=options, args=args):
                list_result = unparser.unparse_options_and_args_to_list(None, args)
                self.assertListEqual(reference_args, list_result)
                result = unparser.unparse_options_and_args(None, args)
                self.assertEqual(' '.join(reference_args), result)

                list_result = unparser.unparse_options_and_args_to_list(options, None)
                self.assertListEqual(reference_options, list_result)
                result = unparser.unparse_options_and_args(options, None)
                self.assertEqual(' '.join(reference_options), result)

                list_result = unparser.unparse_options_and_args_to_list(options, args)
                self.assertListEqual(reference_options + reference_args, list_result)
                result = unparser.unparse_options_and_args(options, args)
                self.assertEqual(
                    ' '.join(itertools.chain(reference_options, reference_args)), result)

    def test_unparse(self):
        unparser = ArgumentUnparser()
        for (reference_options, options), (reference_args, args) in OPTIONS_AND_ARGUMENTS_VARIANTS:
            with self.subTest(options=options, args=args):
                result = unparser.unparse(*args, **options)
                for item in itertools.chain(reference_options, reference_args):
                    self.assertIn(item, result)
                self.assertEqual(
                    ' '.join(itertools.chain(reference_options, reference_args)), result)
                list_result = unparser.unparse_to_list(*args, **options)
                self.assertSetEqual(set(reference_options + reference_args), set(list_result))
                self.assertListEqual(reference_options + reference_args, list_result)

    def test_unparse_nothing(self):
        unparser = ArgumentUnparser()
        _LOG.debug('testing unparsing nothing for: %s', unparser)
        self.assertEqual(unparser.unparse(), '')
        self.assertEqual(unparser.unparse_to_list(), [])



"""Examples for tests of argunparse."""

import collections
import itertools
import pathlib
import typing as t

_UNPARSER_INIT_ARGS_EXAMPLES = [
    ('short_opt', {'-', '--', '_', '__'}),
    ('long_opt', {'-', '--', '_', '__'}),
    ('opt_value', {'=', ' ', ''})]

OPTIONS: dict[str, dict[str, t.Any]] = {
    '-h': {'h': True}, '--verbosity=100': {'verbosity': 100},
    '--DEFINE=NDEBUG': {'DEFINE': 'NDEBUG'},
    '--long_flag': {'long_flag': True}, '-o=out_file.txt': {'o': 'out_file.txt'},
    '--log=log_file.txt': {'log': 'log_file.txt'}, "--what='is that'": {'what': 'is that'},
    '''--extra_args="ok 'well' yeah"''': {'extra_args': "ok 'well' yeah"}}

OPTIONS_VARIANTS = [
    ([ref1, ref2], collections.OrderedDict(itertools.chain(opt1.items(), opt2.items())))
    for ((ref1, opt1), (ref2, opt2)) in itertools.permutations(OPTIONS.items(), 2)]

OPTIONS_SKIPPED = {'skipped': False, 'e': False}

OPTIONS_SKIPPED_VARIANTS = [
    (['-v'], collections.OrderedDict([('v', True), ('skipped', None), ('e', False)])),
    (['--help'], collections.OrderedDict([('skipped', False), ('help', True), ('e', False)])),
    ([], collections.OrderedDict([('skipped', False), ('e', None)]))]

ARGUMENTS = {
    ('my_output_file.txt', 'my_output_file.txt'), ('C:\\Users\\user\\', 'C:\\Users\\user\\'),
    ('/home/user/Desktop', '/home/user/Desktop'), ('0', 0), ('""', ''),
    ('3.1415', 3.1415), ('True', True), ('False', False), ('None', None),
    (str(pathlib.Path('tmp', 'file')), pathlib.Path('tmp', 'file')), ('in_file.txt', 'in_file.txt'),
    ('-123456790', -123456790), (str(2**16), 2**16), ('x', 'x'), ('y', 'y'), ('z', 'z'),
    (repr('hello world'), 'hello world')}

ARGUMENTS_VARIANTS = [([ref1, ref2], [arg1, arg2])
                      for ((ref1, arg1), (ref2, arg2)) in itertools.permutations(ARGUMENTS, 2)]

OPTIONS_AND_ARGUMENTS_VARIANTS = list(itertools.product(OPTIONS_VARIANTS, ARGUMENTS_VARIANTS))
